/**************************************************************************/
/*                                                                        */
/* mpfrx_rem                                                               */
/*                                                                        */
/* computes the remainder of f divided by g                               */
/*                                                                        */
/* Copyright (C) 2007 Andreas Enge                                        */
/*                                                                        */
/* Last changes: August 31, 2007                                          */
/*                                                                        */
/**************************************************************************/

#include "mpfrx-impl.h"

static void mpfrx_rem_naive (mpfrx_ptr r, mpfrx_srcptr f, mpfrx_srcptr g);
static void mpfrx_rem_newton (mpfrx_ptr r, mpfrx_srcptr f, mpfrx_srcptr g);

/*****************************************************************************/

void mpfrx_rem (mpfrx_ptr r, mpfrx_srcptr f, mpfrx_srcptr g)

{
   if (f->deg < g->deg)
      mpfrx_set (r, f);
   else
      if (g->deg >= 32 && f->deg >= 63)
         mpfrx_rem_newton (r, f, g);
      else
         mpfrx_rem_naive (r, f, g);
}

/*****************************************************************************/

static void mpfrx_rem_naive (mpfrx_ptr r, mpfrx_srcptr f, mpfrx_srcptr g)
   /* computes the remainder of f divided by g by the naive algorithm for    */
   /* deg f >= deg g                                                         */
{
   mpfr_t tmp;
   int i, j;

   if (mpfr_cmp_si (g->coeff [g->deg], 1) == 0) {
      mpfr_init2 (tmp, mpfr_get_prec (f->coeff [0]));
      /* the only case we actually use for multievaluation */
      mpfrx_set (r, f);
      for (i = f->deg - g->deg; i >= 0; i--)
         for (j = 0; j < g->deg; j++) {
            mpfr_mul (tmp, r->coeff [g->deg + i], g->coeff [j], GMP_RNDN);
            mpfr_sub (r->coeff [i + j], r->coeff [i + j], tmp, GMP_RNDN);
         }
      r->deg = g->deg - 1;
      /* To save memory, one might want to do a realloc on r here:
      mpfrx_realloc2 (r, r->deg + 1);
      */
   }
   else {
      printf ("*** Houston, we have a problem!\n");
      printf ("*** Calling mpfrx_rem_naive with a non-monic divisor.\n");
      printf ("*** Go back programming!\n");
      exit (1);
   }
}

/*****************************************************************************/

static void mpfrx_rem_newton (mpfrx_ptr r, mpfrx_srcptr f, mpfrx_srcptr g) {
   /* computes the remainder of f divided by g by Newton iterations for      */
   /* deg f >= deg g                                                         */

   mpfrx_t q, tmp;

   mpfrx_init2 (q, f->deg - g->deg + 1, r->prec);
   mpfrx_init2 (tmp, f->deg + 1, r->prec);

   mpfrx_rev (tmp, f, 0);
   mpfrx_rev (q, g, 0);
   mpfrx_sinv (q, q, f->deg - g->deg + 1);
   mpfrx_smul (q, tmp, q, f->deg - g->deg + 1);
   mpfrx_rev (q, q, f->deg - g->deg);

   mpfrx_mul (tmp, q, g);
   mpfrx_sub (r, f, tmp);
   if (r->deg >= g->deg)
      r->deg = g->deg - 1;

   mpfrx_clear (q);
   mpfrx_clear (tmp);
}

